#include "Ukkonen.h"

Ukkonen::Ukkonen(bool countOnly, const char* pattern, size_t m, int error): ApproximateMatcher(countOnly){
	this -> pattern = pattern;
	this -> m = m;
	this -> error = error;
	n = 0;

	tree.push_back(UkkonenTreeNode());
	initialAutomatonState = currentAutomatonState = buildFDA();
}

int Ukkonen::update(char* chunk, size_t chunkSize) {
	int initialPos = (n==0) ? 0 : (m+error-1);

	for(int i = initialPos; i < chunkSize; ++i) {
		if(chunk[i] != '\n') {
			currentAutomatonState = automaton[currentAutomatonState].nextFn[ALPHABET_INDEX(chunk[i])];
			if(automaton[currentAutomatonState].minError <= error) 
				registerAMatching(chunk, i, m, error);
		} else {
			//recomeça o casamento no fim da linha
			currentAutomatonState = initialAutomatonState;
		}
		++n;
	}

	return m+error-1;
}

int Ukkonen::getColumnFromTree(int* col, bool* wasCreated) {
	int cur = 0;

	for(int i = 1; i < m; ++i) {
		int dif = col[i] - col[i-1];

		if(tree[cur].getChild(dif) == -1) {
			tree[cur].setChild(dif, tree.size());
			tree.push_back(UkkonenTreeNode());
			*wasCreated = true;
		}

		cur = tree[cur].getChild(dif);
	}

	/*
	 * Ao invés de inserir uma folha na árvore, vamos inserir um nó 
	 * no autômato. Como a indexação tanto da árvore quanto do 
	 * autômato é feita através de inteiros, não precisamos fazer
	 * essa diferenciação no array de filhos (dos nós da árvore)
	 */
	int dif = col[m] - col[m-1];
	if(tree[cur].getChild(dif) == -1) {
		tree[cur].setChild(dif, automaton.size()); 
		automaton.push_back(UkkonenAutomatonNode(col[m]));
		*wasCreated = true;
	}

	return tree[cur].getChild(dif);
}

void Ukkonen::calculateNextColumn(int* col, int ch, int* nextCol) {
	nextCol[0] = 0;

	for(int i = 1; i <= m; ++i) {
		int phi = (ALPHABET_INDEX(pattern[i-1]) == ch) ? 0 : 1;
		nextCol[i] = MIN(error + 1, MIN(phi + col[i-1], 1 + MIN(col[i], nextCol[i-1])));
	}
}

int Ukkonen::buildFDA() {
	queue<pair<int, int*> > queue;
	MemPool pool(m);
	int *firstCol = new int[m+1];
		
	//caso base, primeira coluna
	for(int i = 0; i <= m; ++i)
		firstCol[i] = i;

	bool wasCreated;
	//insere na árvore essa coluna
	int firstNode = getColumnFromTree(firstCol, &wasCreated);
	queue.push(make_pair(firstNode, firstCol));

	while(!queue.empty()) {
		int currentNode = queue.front().first;
		int* currentCol = queue.front().second;
		queue.pop();

		for(int i = 0; i < ALPHABET_SIZE; ++i) {
			int* nextCol = pool.borrow();
			calculateNextColumn(currentCol, i, nextCol);
			
			bool wasCreated = false; //A coluna não estava representada, mas foi criada
			int nextNode = getColumnFromTree(nextCol, &wasCreated);

			if(wasCreated) {
				queue.push(make_pair(nextNode, nextCol));
			} else 
				pool.giveBack(nextCol);

			automaton[currentNode].nextFn[i] = nextNode;
		}
		pool.giveBack(currentCol);
	}

	return firstNode;
}