#include <getopt.h>
#include <cstdlib>
#include <cstdio>

#include <string>
#include <vector>
using std::string;
using std::vector;

struct PmtConfiguration {
	//O usuário pode escolher qual algoritmo utilizar
	string algorithmName;

	//Arquivo contendo os padrões
	string patternFileName;

	//Um único padrão
	string pattern;

	//O usuário pode visualizar os alinhamentos encontrados (casamento aproximado)
	int showAlignmentFlag; 

	//Apenas contar os casamentos encontrados
	int countOnlyFlag;

	//Exibe a mensagem de ajuda
	int showHelpMessageFlag;

	//O erro máximo para o casamento aproximado
	int maxError;

	//Os nomes dos arquivos onde os padrões serão procurados
	vector<string> textFiles;

	PmtConfiguration() {
		algorithmName = patternFileName = patternFileName = "";
		showHelpMessageFlag = countOnlyFlag = showAlignmentFlag = 0;
		maxError = -1;
	}

	bool validateConfig() {
		if(showHelpMessageFlag) {
			return true;
		}

		bool isExactMatching = (maxError == -1);
		bool valid = true;

		//Valida a escolha do algoritmo
		if(algorithmName != "") {
			if(isExactMatching 
				&& algorithmName != "k"
				&& algorithmName != "b"
				&& algorithmName != "sh"
				&& algorithmName != "a"){
				printf("Erro: o algoritmo escolhido (%s) para casamento exato é inválido!\n", algorithmName.c_str());
				printf("\tAs opções são [k]mp, [b]oyer moore, [a]ho-corasick [sh]ift-or\n");
				valid = false;
			} else if(isExactMatching == false
				&& algorithmName != "se"
				&& algorithmName != "u"
				&& algorithmName != "w") {
				printf("Erro: o algoritmo escolhido (%s) para casamento aproximado é inválido!\n", algorithmName.c_str());
				printf("\tAs opções são [se]llers, [u]kkonen e [w]u-mamber\n");
				valid = false;
			} 
		}

		//Confere se o usuário forneceu uma fonte de padrões
		if(pattern == "" && patternFileName == "") {
			printf("Erro: informe uma fonte de padrões. Veja os exemplos abaixo:\n");
			printf("\tpmt --edit 10 --pattern pats.txt\n");
			printf("\tpmt pattern text\n");
			valid = false;
		}

		//Verifica se a opção de reconstrução de alinhamento foi definida de forma consistente
		if(showAlignmentFlag && isExactMatching) {
			printf("Erro [-s, --show_alignment]: a opção para reconstrução de alinhamento só é válida para casamentos aproximados.\n");
			valid = false;
		}

		//Verifica se a reconstrução de alinhamento é suportada pelo algoritmo escolhido
		if(showAlignmentFlag && algorithmName != "" && algorithmName != "se") {
			printf("Erro [-s, --showAlignmentFlag e -a, --algorithmName = (%s)]: A reconstrução de alinhamento só é suportada pelo algorithm Sellers.\n", algorithmName.c_str());
			valid = false;
		}

		//Confere se o usuário forneceu uma fonte de texto
		if(textFiles.empty()) {
			printf("Erro: nenhum arquivo de texto foi especificado. Forneça os arquivos como último argumento. Veja os exemplos:\n");
			printf("\tpmt --edit 10 --pattern pats.txt  text.txt\n");
			printf("\tpmt --pattern pats.txt  texto.txt text_sample_*.txt\n");
			valid = false;
		}

		return valid;
	}

} config;

/*
* Configuração para a função getopt_long.
* Adaptado de 
* http://www.gnu.org/software/libc/manual/html_node/Getopt-Long-Option-Example.html
*/
struct option options[] =
{
	{"show-alignment", no_argument, &config.showAlignmentFlag, 1},
	{"count-only",   no_argument, &config.countOnlyFlag, 1},
	{"help",   no_argument, &config.showHelpMessageFlag, 1},
	{"algorithm", required_argument, 0, 'a'},
	{"edit", required_argument, 0, 'e'},
	{"pattern", required_argument, 0, 'p'},
	{0, 0, 0, 0}
};

/*
* Análise dos parâmetros do programa (preenchimento dos campos de 'config')
* Adaptado de 
* * http://www.gnu.org/software/libc/manual/html_node/Getopt-Long-Option-Example.html
*/
PmtConfiguration& parseOptions(int argc, char* argv[]) {
	
	while(true){
		/*
		* -a x = --algorithm x
		* -e x = --edit x
		* -p x = --pattern x
		* -s = --show_alignment
		* -c = --count_only
		* -h = --help
		*/
	    int c = getopt_long (argc, argv, "a:e:p:sch", options, NULL);

	    if(c == -1) 
	    	break;

	    if(c == 0) {//Define uma flag, não precisa ser tratado aqui
	    	continue;
	    } else if(c == 'a') {
		  config.algorithmName = optarg;
		} else if(c == 'e') {
		  config.maxError = atoi(optarg);
		} else if(c == 'p') {
		  config.patternFileName = optarg;
		} else if(c == 's') {
		  config.showAlignmentFlag = 1;
		} else if(c == 'c') {
		  config.countOnlyFlag = 1;
		}else if(c == 'h') {
			config.showHelpMessageFlag = 1;
		} else {
			exit(1);
		}
	}

	//parâmetros sem opções. Podem ser o padrão e os arquivos de texto
	if (optind < argc) {
		//Usuário não informou uma fonte de padrões.
		//Então a primeira string é um padrão.
		if(config.patternFileName == "")
			config.pattern = argv[optind++];

		//As demais strings são arquivos que fornecem o texto
		while(optind < argc)
			config.textFiles.push_back(argv[optind++]);
	}

	return config;
}

void printHelpMessage() {
	printf("\n\nUso: pmt [options] pattern textfile [textfile...]\n\n\n");

	printf("Opções:\n");
	printf("\t-p FILE_NAME, --pattern FILE_NAME:\tRealiza a busca de todos os padrões contidos no arquivo FILE_NAME.\n");
	printf("\t-e MAX_ERROR, --edit MAX_ERROR:   \tLocaliza todas as ocorrencias aproximadas do padrão a uma distância de edição maxima MAX_ERROR.\n");
	printf("\t-a , --algorithm [u|se|w|k|b|sh|a]:   \tUtiliza o algoritmo especificado para realizar o casamento.\n");
	printf("\t                                       \tOs algoritmos disponíveis para casamento APROXIMADO são:\n");
	printf("\t                                       \t\t[u]kkonen, [se]llers e [w]u-mamber.\n");
	printf("\t                                       \tOBS: o algoritmo wu-mamber suporta apenas padrões com tamanho até 64.\n");
	printf("\t                                       \tOs algoritmos disponíveis para casamento EXATO são:\n");
	printf("\t                                       \t\t[k]mp, [b]oyer-moore, [sh]ift-or e [a]ho-corasick.\n");
	printf("\t                                       \tOBS: Caso mais de um padrão seja fornecido (por um arquivo), apenas o Aho Corasick pode ser utilizado.\n");
	printf("\t-s , --show-alignment:            \tMostra os alinhamentos encontrados no casamento aproximado.\n");
	printf("\t                                  \tOBS: suportado apenas pelo algoritmo sellers.\n");
	printf("\t-c , --count-only:                \tImprime apenas a quantidade de ocorrências encontradas.\n");

	printf("\n\n");
	printf("As únicas informações obrigatórias são (1) uma fonte de padrões e (2) uma fonte de texto.\n");
	printf("Em particular, o algoritmo a ser utilizado no casamento é, por padrão, automaticamente escolhido de acordo\n");
	printf("com parâmetros variados, como o tamanho do padrão e do alfabeto.\n");
	printf("A utilização mais simples da ferramenta:\n\n");
	printf("\t./pmt pattern text-file.txt\n\n");
	printf("procura e imprime as ocorrências exatas do padrão 'pattern' no arquivo 'text-file.txt'.\n");
	printf("Outras exemplos de utilização são:\n\n");
	printf("\t./pmt -c pattern text-file.txt\t//apenas conta quantas ocorrências foram encontradas e não as imprime\n\n");
	printf("\t./pmt -f pattern-file text-file.txt\t//procura por ocorrências de qualquer um dos padrões em \'pattern-file\'\n");
}
