#include "SellersRec.h"
#include "Sellers.h"
#include "WuMamber2.h"
#include "Ukkonen.h"

#include "ExactMatcher.h"
#include "Kmp.h"
#include "BoyerMoore.h"
#include "ShiftOr.h"
#include "ShiftOr2.h"
#include "BruteForce.h"
#include "Aho.h"

#include "OptionsParsing.cpp"

#include <string>
#include <vector>
#include <cstdio>
using std::string;
using std::vector;

#define UKKONEN_TEXT_SIZE_TRESHOLD (100*(1<<20))

const size_t MAX_BUFFER_SIZE = 1<<24;
char buffer[MAX_BUFFER_SIZE+1];

void presentFileToAlgorithm(const char* fileName, Matcher* algorithm) {
	if(FILE* fText = fopen(fileName, "r")){
		size_t currentBufferSize = 0, toCopy = 0;
		while(1){
			for(int i = 0; i < toCopy; ++i)
				buffer[i] = buffer[currentBufferSize - toCopy + i];

			currentBufferSize = fread(buffer+toCopy, sizeof(char), MAX_BUFFER_SIZE - toCopy, fText);
			currentBufferSize += toCopy;

			buffer[currentBufferSize] = 0;
			toCopy = (size_t) algorithm->update(buffer, currentBufferSize);

			if(currentBufferSize != MAX_BUFFER_SIZE) break;
		}

		fclose(fText);
	}
}

long int getFileSize(const char* fileName) {
	long int size = 0;

	if(FILE* file = fopen(fileName, "rb")){
		fseek(file, 0L, SEEK_END);
		size = ftell(file);
		fclose(file);
	}

	return size;
}

Matcher* chooseAlgorithmForApproximateMatching(const char* pattern, size_t m, ull totalTextSize, PmtConfiguration& config) {
	int error = config.maxError;
	bool countOnly = config.countOnlyFlag;

	if(config.showAlignmentFlag) //Apenas a implementação do Sellers suporta essa opção
		return new SellersRec(pattern, m, error);

	if(config.algorithmName ==  "se")  //O usuário especificou o Sellers
		return new Sellers(countOnly, pattern, m, error);

	if(config.algorithmName == "u") // O usuário especificou o Ukkonen
		return new Ukkonen(countOnly, pattern, m, error);

	if(config.algorithmName == "w") //O usuário especificiou o Wu-Mamber
		return new WuMamber2(countOnly, pattern, m, error);
		
	/* 
	 * ===============================================================================================================
	 * 								Caso o usuário não tenha especificado o algoritmo:
	 * ===============================================================================================================
	 * A estratégia geral é utilizar o Ukkonen para arquivos maiores que 100MB
	 * sempre que isso for 'seguro', isto é, quando o autômato não crescer muito.
	 * Para julgar isso consideramos a seguinte heurística:
	 * 	(i) m < 20 : a quantidade de estados é limitada por 3^m, que é aceitável para m < 20. Na prática não chega a
	 *      tantos estados.
	 *  (ii) 20 <= m < 70 e error <= 5: O tamanho do padrão é moderadamente grande, mas esperamos que o
	 * 		erro pequeno ajude a conter o crescimento do autômato.
	 *  (iii) 70 <= m < 100 e error <= 3: O padrão é grande, mas esperamos que o erro pequeno ajude a 
	 *		conter o crescimento do autômato.
	 *
	 * Caso não seja 'seguro' usar o Ukkonen, teremos que escolher entre o Sellers e o WuMamber. A implementação do 
	 * WuMamber só suporta padrões com m <= 64. Através de experimentos, verificamos que o WuMamber é quase 2x mais 
	 * rápido que o Sellers para o mesmo parâmetro ('m' para o Sellers e 'error' para o WuMamber). Por isso, usaremos 
	 * o WuMamber sempre que m >= error/2.
	 */

	//Se for vantajoso e 'seguro' usar o Ukkonen
	if(totalTextSize >= UKKONEN_TEXT_SIZE_TRESHOLD 
			&& (m < 20 || (m < 70 && error <= 5) || (m < 100 && error <= 3)))
		return new Ukkonen(countOnly, pattern, m, error);

	if(m > error / 2 && m <= 64) //Escolhemos o WuMamber
		return new WuMamber2(countOnly, pattern, m, error);

	//Quando todo o resto falhar, escolhemos o Sellers
	return new Sellers(countOnly, pattern, m, error);
}


Matcher* chooseAlgorithmForExactMatching(const char* pattern, size_t m, PmtConfiguration& config) {
	if(config.algorithmName == "k") //O usuário especificou o Kmp
		return new Kmp(config.countOnlyFlag, pattern, m);

	if(config.algorithmName == "b") //O usuário especificou o BoyerMoore
		return new BoyerMoore(config.countOnlyFlag, pattern, m);

	/*
	 * O usuário especificou o ShiftOr. Nesse caso existem duas oções:
	 *	(i) m <= 64: Esse caso é suportado pela implementação ShiftOr2, que é mais eficiente
	 *  (ii) m > 64: Caso suportado apenas pela implementação ShiftOr
	 */
	if(config.algorithmName == "sh") {
		if(m <= 64) 
			return new ShiftOr2(config.countOnlyFlag, pattern, m);
		return new ShiftOr(config.countOnlyFlag, pattern, m);
	}

	/* 
	 * ===============================================================================================================
	 * 								Caso o usuário não tenha especificado o algoritmo:
	 * ===============================================================================================================
	 * A estratégia geral é utilizar o Boyer Moore sempre que isso for vantajoso. De maneira geral, o Boyer Moore
	 * tem problemas de eficiência quando o tamanho do padrão é pequeno e quando muitos casamentos parciais (grandes)
	 * ocorrem (em consequência do tamanho do alfabeto e/ou das próprias características do padrão e do texto).
	 * Para os casos de padrões pequenos, escolhemos entre a força bruta (m=1) ou o ShiftOr2 (m <= 5). Para m > 5
	 * consideramos o Boyer Moore como algoritmo padrão. No entanto, quando o tamanho do alfabeto for pequeno,
	 * utilizamos o ShiftOr2, caso possível.
	 */

	if(m == 1)
		return new BruteForce(config.countOnlyFlag, pattern, m);
	
	if(m <= 5) 
		return new ShiftOr2(config.countOnlyFlag, pattern, m);
	
	BoyerMoore* moore = new BoyerMoore(config.countOnlyFlag, pattern, m);

	if(ALPHABET_SIZE <= 5 && m <= 64) { //Desiste do Boyer Moore
		delete moore;

		return new ShiftOr2(config.countOnlyFlag, pattern, m);
	}

	return moore;
}

int main(int argc, char* argv[]){
	PmtConfiguration& config = parseOptions(argc, argv);

	if(!config.validateConfig()){
		printf("\n\nUse a opção -h, --help para obter mais informações.\n");
		exit(1);
	}
	
	if(config.showHelpMessageFlag) {
		printHelpMessage();
		exit(1);
	}


	//Para facilitar a leitura do arquivo de padrões
	vector<string> patterns;
	
	//Para facilitar a passagem dos padrões para os algoritmos
	vector<const char*> patternsCStyle; 

	//padrões vêm de um arquivo
	if(config.patternFileName != "") {
		if(FILE* pats = fopen(config.patternFileName.c_str(), "r")) {
			while(fgets(buffer, sizeof buffer, pats)) {
				int size = strlen(buffer);
				if(buffer[size-1] == '\n') { //remove a quebra de linha do final
					buffer[size-1] = 0;
				}
				patterns.push_back(buffer);
				patternsCStyle.push_back(patterns.back().c_str());
			}

			fclose(pats);
		}
	} else { //padrão vem da linha de comando
		patterns.push_back(config.pattern);
		patternsCStyle.push_back(config.pattern.c_str());
	}


	/*
	 * Se o usuário escolheu o Aho Corasick ou não especificou nenhum algoritmo,
	 * existem vários padrões a serem procurados e o casamento é exato, vamos
	 * utilizar o Aho Corasick.
	 */
	if(config.algorithmName == "a" || (config.maxError == -1 && config.algorithmName == "" && patternsCStyle.size() > 1)) {
		for(int i = 0; i < patternsCStyle.size(); ++i){
			Alphabet::incrementalSetup(patternsCStyle[i], patterns[i].size());
		}

		Aho* aho = new Aho(config.countOnlyFlag, patternsCStyle);

		for(int t = 0; t < config.textFiles.size(); ++t) {
			aho->resetForNewFile();
			printf("Ocorrências dos padrões no arquivo \'%s\':\n", config.textFiles[t].c_str());
			presentFileToAlgorithm(config.textFiles[t].c_str(), aho);
			printf("%lld ocorrências no total.\n\n", aho->getNumberOfMatchings());
		}
	} else {
		/*
		 * Nesse caso, escolhemos um algoritmo para cada padrão da entrada.
		 */


		/*
		 * Cálculo da soma do tamanho dos arquivos de texto. Queremos apenas saber se essa soma é maior que 100MB, por
		 * isso paramos o laço assim que esse valor é alcançado. Esse valor é ssado apenas caso o casamento seja aproximado.
		 */
		ull totalTextSize = 0;
		if(config.maxError != -1) {
			for(int i = 0; i < config.textFiles.size() && totalTextSize <= UKKONEN_TEXT_SIZE_TRESHOLD; ++i) {
				totalTextSize += getFileSize(config.textFiles[i].c_str());
			}
		}

		/*
		 * Procura cada padrão em todos os arquivos de texto.
		 */
		for(int i = 0; i < patternsCStyle.size(); ++i) {
		 	Matcher* algorithm = NULL; //O algoritmo escolhido

			size_t m = patterns[i].size(); // Tamanho do padrão

			Alphabet::setup(patternsCStyle[i], m); //Configura o alfabeto para conter apenas os caracteres do padrão

			//Casamento aproximado
			if(config.maxError != -1) 
				algorithm = chooseAlgorithmForApproximateMatching(patternsCStyle[i], m, totalTextSize, config);

			else //Caso exato 
				algorithm = chooseAlgorithmForExactMatching(patternsCStyle[i], m, config);

			//Apresentamos cada arquivo de texto ao padrão
			for(int t = 0; t < config.textFiles.size(); ++t) {
				algorithm -> resetForNewFile();
				printf("Ocorrências do padrão \'%s\' no arquivo \'%s\' (busca utilizando o algoritmo %s).\n", patternsCStyle[i], config.textFiles[t].c_str(), algorithm->getAlgoName());
				presentFileToAlgorithm(config.textFiles[t].c_str(), algorithm);
				printf("%llu ocorrências foram encontradas.\n", algorithm->getNumberOfMatchings());
			}

			delete algorithm;
		}

	}
}
